<?php

declare(strict_types=1);

namespace Drupal\entity_sync_mailgun\EntitySync\Api;

// Drupal modules.
use Drupal\entity_sync\Client\ClientInterface;
use Drupal\entity_sync\Exception\InvalidConfigurationException;
// Third-party libraries.
use KrystalCode\ApiIterator\ClientInterface as IteratorClientInterface;
use KrystalCode\ApiIterator\Iterator as ApiIterator;
use Mailgun\Exception\HttpClientException;
use Mailgun\Mailgun;
use Mailgun\Message\MessageBuilder;
use Mailgun\Model\Event\Event;
use Mailgun\Model\Event\EventResponse;
use Mailgun\Model\Message\ShowResponse;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;
use Symfony\Component\Serializer\Serializer;

/**
 * The API client Entity Sync adapter for the Mailgun message v3 APIs.
 */
class MessageClient implements ClientInterface, IteratorClientInterface {

  /**
   * Constructs a new MessageClient object.
   *
   * @param string $apiKey
   *   The Mailgun API key.
   * @param string $domain
   *   The Mailgun domain.
   * @param string $endpoint
   *   The Mailgun endpoint, or `NULL` to use the default endpoint as defined by
   *   Mailgun.
   */
  public function __construct(
    protected string $apiKey,
    protected string $domain,
    protected string|null $endpoint = NULL,
  ) {
  }

  /**
   * {@inheritdoc}
   *
   * Additional supported options are:
   * - event: (string, optional) The type of events to retrieve.
   *
   * Mailgun makes available different types of events not only when an email is
   * incoming but also, for example, when an email has been delivered or when a
   * recipient opens the email. The even type that we want for importing
   * incoming emails is `stored`.
   *
   * Note that events/messages are not available through the APIs immediately
   * after the time they were received by Mailgun because there is a delay until
   * they reach storage through their network of servers. Mailgun therefore
   * advises to implement a specific technique to check whether the results miss
   * items, and wait for at least 30 seconds before repeating the request in
   * that case. We do not implement this though because it complicates a lot of
   * things especially in the context of the Drupal cron job system that
   * commonly runs the imports. Instead, the caller of the import should request
   * to receive messages of up to, say, 5 minutes before the current time. That
   * is an easy way to ensure that the response will contain all messages since
   * that will be enough time for them to reach storage in Mailgun.
   *
   * @link https://documentation.mailgun.com/docs/mailgun/user-manual/events/#event-polling
   */
  public function importList(array $filters = [], array $options = []) {
    return $this->list(
      $this->buildListRequestOptions($filters, $options),
      [
        'begin' => $filters['changed_start'] ?? NULL,
        'end' => $filters['changed_end'] ?? NULL,
        'event' => $options['event'] ?? 'stored',
      ]
    );
  }

  /**
   * {@inheritdoc}
   */
  public function importEntity($id) {
    throw new \Exception(
      'Importing individual Mailgun messages is not supported yet.'
    );
  }

  /**
   * {@inheritdoc}
   */
  public function create(array $fields, array $options = []) {
    $builder = $this->getClient()
      ->messages()
      ->getBatchMessage($this->domain);

    [$fields, $variables] = $this->collectTemplateVariables($fields);
    foreach ($fields as $field_name => $field_value) {
      if ($field_value === NULL) {
        continue;
      }

      $this->buildMessageField($builder, $field_name, $field_value);
    }

    foreach ($variables as $key => $value) {
      $builder->addTemplateVariable($key, $value);
    }

    $builder->finalize();
    $message_ids = $builder->getMessageIds();

    // We provide the message ID in the response as it may be required by
    // subscribers, for example, so that it is stored in a Drupal entity
    // field. It may be used in determining whether a message has been sent
    // and/or to be able to add the reply-to header in a response.
    $message = new \stdClass();
    $message->id = current($message_ids);
    return $message;
  }

  /**
   * {@inheritdoc}
   */
  public function update($id, array $fields, array $options = []) {
    throw new \Exception('Updating Mailgun messages is not supported.');
  }

  /**
   * {@inheritdoc}
   */
  public function list(array $options = [], array $query = []) {
    $options = array_merge(
      [
        'page' => 1,
        'limit' => 100,
        'delay' => NULL,
      ],
      $options
    );
    $query = array_merge(
      [],
      $query
    );
    $query['offset'] = ($options['page'] - 1) * $options['limit'];
    $query['limit'] = $options['limit'];

    // By default, we return an iterator. Caller code can then loop over it
    // and the iterator will internally handle making the API calls for
    // fetching the objects for each page.
    if (empty($options['bypass_iterator'])) {
      return new ApiIterator(
        $this,
        $options['page'],
        $options['limit'],
        $query,
        FALSE,
        $options['delay']
      );
    }

    // Otherwise, either it is requested to fetch the results directly, or
    // this is a call by the iterator while being looped over. Make the
    // request and return the results.
    unset($options['bypass_iterator']);

    // Unfortunately, Mailgun does not make the list of messages available
    // directly. We first need to get the list of events and then fetch the
    // message for each event, one by one.
    $event_response = $this->getEvents($query);
    $query['paging'] = [
      'next' => $event_response->getNextUrl(),
      'previous' => $event_response->getPreviousUrl(),
      'first' => $event_response->getFirstUrl(),
      'last' => $event_response->getLastUrl(),
    ];
    $events = $event_response->getItems();

    // We are not able to determine whether there are more results from the
    // above paging information included in the response because Mailgun always
    // provides the URL to the next page even when we are on the last page.
    $has_more = NULL;
    $count = count($events);
    if ($count === 0 || $count < $options['limit']) {
      $has_more = FALSE;
    }

    // Fetch the message and convert the response to the format expected by
    // Entity Synchronization.
    // phpcs:disable
    // @I Respect any delay settings between API calls
    //    type     : bug
    //    priority : low
    //    labels   : import, message
    //    notes    : Mailgun documents API rate limits for the number of emails
    //               being sent, but not for the number API calls that can be
    //               made - this might not be a problem. That should be
    //               confirmed though and it should be implemented anyways for
    //               respecting the provided setting.
    // phpcs:enable
    $messages = array_filter(array_map([$this, 'getMessage'], $events));
    $messages = array_map([$this, 'serializeMessage'], $messages);

    return [
      new \CachingIterator(
        new \ArrayIterator($messages),
        \CachingIterator::FULL_CACHE
      ),
      $has_more,
      $query,
    ];
  }

  /**
   * Separates fields from template variables.
   *
   * Template variables can be passed by using `'%template_variable%_'` as a
   * prefix in the remote name of a field in the field mapping.
   *
   * @param array $fields
   *   An associative array containing the values for the fields being exported,
   *   keyed by the remote field name.
   *
   * @return array
   *   A numerical array containing the following elements in the given order:
   *   - The values for the actual remote message fields, keyed by the remote
   *     field name.
   *   - The values for the template variables, keyed by the variable name
   *     without the `'%template_variable%_'` prefix.
   */
  protected function collectTemplateVariables(
    array $fields,
  ): array {
    $prefix = '%template_variable%_';
    $prefix_length = strlen($prefix);

    // Separate variables from fields.
    $variables = \array_filter(
      $fields,
      fn ($value, $key) => \str_starts_with($key, $prefix),
      ARRAY_FILTER_USE_BOTH
    );
    $fields = array_diff_key($fields, $variables);

    // Remove the prefix from variables.
    $variables = \array_combine(
      \array_map(
        fn ($variable) => \substr($variable, $prefix_length),
        \array_keys($variables)
      ),
      $variables
    );

    return [$fields, $variables];
  }

  /**
   * Builds the message for the given field.
   *
   * Each method provided by the Mailgun Message API client for setting the
   * value of a message property requires a different number of arguments. We
   * therefore require the field value to be always provided as an array which
   * we then unpack when calling the client method. Otherwise we'd need to add
   * support for each method one by one.
   *
   * Methods for message properties that allow multiple values can be
   * called multiple times. The values for those are expected to be contained
   * within an extra array wrapper i.e. an array containing arrays of the
   * arguments that the method expects.
   *
   * The field names are expected to be provided in snake case without the
   * `add/set` prefix of the corresponding camel case method provided by the
   * Mailgun Message API client. For example, use `to_recipient` to provide the
   * arguments for the `addToRecipient` method.
   *
   * @param \Mailgun\Message\MessageBuilder $builder
   *   The message builder.
   * @param string $field_name
   *   The name of the field.
   * @param array $field_value
   *   An array containing the value of the field.
   *
   * @see \Mailgun\Message\MessageBuilder
   */
  protected function buildMessageField(
    MessageBuilder $builder,
    string $field_name,
    array $field_value,
  ): void {
    $field_parts = explode('_', $field_name);
    $field_parts = array_map('ucfirst', $field_parts);
    $method_main_part = implode('', $field_parts);

    // We must have either a method starting with `add` or with `set` provided
    // by the Mailgun Message API client.
    if (method_exists($builder, "set{$method_main_part}")) {
      $method = "set{$method_main_part}";
      $builder->{$method}(...$field_value);
    }
    elseif (method_exists($builder, "add{$method_main_part}")) {
      foreach ($field_value as $value_item) {
        $method = "add{$method_main_part}";
        $builder->{$method}(...$value_item);
      }
    }
    else {
      throw new InvalidConfigurationException(sprintf(
        'Could not find the Mailgun Message API client method to call for field "%s".',
        $field_name,
      ));
    }
  }

  /**
   * Returns an instantiated Mailgun client.
   *
   * @return \Mailgun\Mailgun
   *   The Mailgun client.
   */
  protected function getClient(): Mailgun {
    $arguments = [$this->apiKey];
    if ($this->endpoint !== NULL) {
      $arguments[] = $this->endpoint;
    }
    return Mailgun::create(...$arguments);
  }

  /**
   * Fetches and returns the events from Mailgun.
   *
   * @param array $query
   *   The query parameters for fetching the objects.
   *
   * @return \Mailgun\Model\Event\EventResponse
   *   The response containing the events.
   */
  protected function getEvents(array $query): EventResponse {
    $parameters = [];
    if (isset($query['begin'])) {
      $parameters['begin'] = $query['begin'];
    }
    if (isset($query['end'])) {
      $parameters['end'] = $query['end'];
    }
    if (isset($query['event'])) {
      $parameters['event'] = $query['event'];
    }

    return $this->getClient()
      ->events()
      ->get($this->domain, $parameters);
  }

  /**
   * Fetches and returns the message for the given event from Mailgun.
   *
   * @param \Mailgun\Model\Event\Event $event
   *   The event.
   *
   * @return \Mailgun\Model\Message\ShowResponse|null
   *   The message, or `NULL` if the event does not contain a message or if the
   *   message was not found.
   */
  protected function getMessage(Event $event): ShowResponse|null {
    $url = $event->getStorage()['url'] ?? NULL;
    if ($url === NULL) {
      return NULL;
    }

    try {
      return $this->getClient()->messages()->show($url);
    }
    catch (HttpClientException $exception) {
      if ($exception->getCode() === 404) {
        return NULL;
      }

      throw $exception;
    }
  }

  /**
   * Converts the message to the format required by Entity Synchronization.
   *
   * That is a simple `stdClass` that has all the message properties exposed as
   * public properties.
   *
   * @param \Mailgun\Model\Message\ShowResponse $message
   *   The message.
   *
   * @return \stdClass
   *   The converted message object.
   */
  protected function serializeMessage(ShowResponse $message): object {
    $serializer = new Serializer(
      [new ObjectNormalizer()],
      [new JsonEncoder()]
    );
    $normalized_message = json_decode(
      $serializer->serialize($message, JsonEncoder::FORMAT)
    );

    // The message ID is only provided within the headers array. We need to have
    // it available as a top-level property as well because it is expected for
    // certain functionality to work, for example, for storing it in a Drupal
    // entity field.
    $header = array_filter(
      $message->getMessageHeaders(),
      fn($h) => $h[0] === 'Message-Id'
    );
    $normalized_message->id = current($header)[1];
    return $normalized_message;
  }

  /**
   * Builds and returns the options for the API iterator client.
   *
   * @param array $filters
   *   The filters passed to the client.
   * @param array $options
   *   The options passed to the client.
   *
   * @return array
   *   The options array to be passed to the iterator client for the list
   *   request.
   */
  protected function buildListRequestOptions(
    array $filters,
    array $options,
  ): array {
    // Support delay between requests that can reduce the risk of hitting
    // API rate and processing limits. If not given, no delay will be applied.
    $request_options = [
      'delay' => $options['delay'] ?? NULL,
    ];

    // Number of items per page. If not given, the iterator client's `list`
    // method will apply the default of 100.
    if (isset($options['limit'])) {
      $request_options['limit'] = $options['limit'];
    }
    // Page to start at; we may be continuing a state-managed import that did
    // not finish all pages in its last run.
    // If not given, the iterator client's `list` method will apply the default
    // of 1 i.e. start at the first page.
    if (!empty($filters['offset']) && isset($request_options['limit'])) {
      $request_options['page'] = ((int) floor($filters['offset'] / $request_options['limit'])) + 1;
    }

    return $request_options;
  }

}
