<?php

declare(strict_types=1);

namespace Drupal\entity_sync_mailgun\Plugin\EntitySync\FieldTransformer;

// Drupal modules.
use Drupal\entity_sync\Exception\FieldImportException;
use Drupal\entity_sync\Exception\InvalidConfigurationException;
use Drupal\entity_sync\FieldTransformer\PluginBase;
use Drupal\file\FileInterface;
use Drupal\file\FileRepositoryInterface;
// Drupal core.
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Component\Render\PlainTextOutput;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\File\Exception\DirectoryNotReadyException;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Utility\Token;
// Third-party libraries.
use Mailgun\Exception\HttpClientException;
use Mailgun\Mailgun;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\Validator\ValidatorInterface;

/**
 * Transformer that imports attachments into file entities.
 *
 * Supported configuration properties:
 * - content_types: (array, optional) An array containing containing the
 *   content types to permit in attachments. If not provided, attachments of all
 *   content types will be permitted to be imported. Supported values are:
 *   - includes (array): An array containing the content types to allow. When
 *     provided, only attachments of the given content types will be imported;
 *     the rest will be skipped.
 *   - excludes (array): An array containing the content types to disallow. When
 *     provided, only attachments that do not have one of the given content
 *     types will be imported; the rest will be skipped.
 * - skip_missing_attachments: (boolean, optional, defaults to `TRUE`) Mailgun
 *   stores the attachments for a limited number of time and trying to retrieve
 *   them after that results in a `404 Not Found` response. This setting
 *   determines the behavior in such cases. When set to `TRUE` and an attachment
 *   is not found, it is skipped without raising an error i.e. it is treated as
 *   if it did not exist allowing the rest of the import to continue. When set
 *   to `FALSE`, an error is raised that might block the import of the entity.
 * - skip_failed_attachments: (boolean, optional, default to `TRUE`) The file
 *   entities created as a result of importing attachments are validated using
 *   the entity validation API. When validation errors exist and this setting is
 *   set to `TRUE`, the errors are ignored and the import proceeds. When set to
 *   `FALSE`, an error will be raised. Note that the file entity is created and
 *   saved in all cases anyways because that's done already already by
 *   `\Drupal\file\FileRepositoryInterface::writeData()`.
 * - format: (string, optional, defaults to `entity_reference`) The format of
 *   the transformed value. Supported values are:
 *   - entity_reference: An array containing arrays with the IDs of the files
 *     created for the attachments, keyed by `target_id` so that they can be
 *     stored as values in entity reference field. For example:
 *     `[['target_id' => '1'], ['target_id' => '2']]`.
 *   - content_ids: An array containing arrays with the content IDs of the
 *     attachments, keyed by the IDs of the files created for them. If the
 *     content ID for an attachment could not be determined, the array item
 *     value will be `FALSE`. For example:
 *     `[['1' => 'f_lyw4u2352'], ['2' => FALSE]]`.
 * - storage: (array, optional) Options related to the storage where the files
 *   will be saved in. Supported options are:
 *   - source: (string, optional, defaults to `target_field`). The source of
 *     the settings that will determine the URI scheme and the directory.
 *     Supported values:
 *     - target_field: The definition of the target field that will reference
 *       the files.
 *     - config: The definition of a field that is identified by plugin
 *       configuration properties.
 *   - entity_type_id: (string, optional) Required if the `config` option is
 *     given; the ID of the entity type of the parent entity of the field.
 *   - bundle: (string, optional) Required if the `config` option is given; the
 *     bundle of the parent entity of the field.
 *   - field_name: (string, optional) Required if the `config` option is given;
 *     the machine name of the field.
 *
 * @EntitySyncFieldTransformer(
 *   id = "mailgun_file_attachments"
 * )
 */
class FileAttachments extends PluginBase {

  /**
   * Constructs a new FileAttachments object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Symfony\Component\Validator\Validator\ValidatorInterface $constraint_validator
   *   The constraint validator.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $constraint_manager
   *   The constraint plugin manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityFieldManager
   *   The entity field manager.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system.
   * @param \Drupal\file\FileRepositoryInterface $fileRepository
   *   The file repository.
   * @param \Drupal\Core\Utility\Token $token
   *   The token service.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    ValidatorInterface $constraint_validator,
    PluginManagerInterface $constraint_manager,
    protected EntityFieldManagerInterface $entityFieldManager,
    protected FileSystemInterface $fileSystem,
    protected FileRepositoryInterface $fileRepository,
    protected Token $token,
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $constraint_validator,
      $constraint_manager
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    ValidatorInterface $constraint_validator,
    array $configuration,
    $plugin_id,
    $plugin_definition,
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $constraint_validator,
      $container->get('plugin.manager.entity_sync_validation_constraint'),
      $container->get('entity_field.manager'),
      $container->get('file_system'),
      $container->get('file.repository'),
      $container->get('token'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'content_types' => NULL,
      'skip_missing_attachments' => TRUE,
      'skip_failed_attachments' => TRUE,
      'format' => 'entity_reference',
      'storage' => ['source' => 'target_field'],
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  protected function validateConfiguration() {
    $includes = $this->configuration['content_types']['include'] ?? NULL;
    $excludes = $this->configuration['content_types']['exclude'] ?? NULL;
    if ($includes !== NULL && $excludes !== NULL) {
      throw new InvalidConfigurationException(
        'You may provide attachment content type includes, excludes, but not both.'
      );
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function transformImportedValue(
    $attachments,
    \stdClass $remote_entity,
    ?ContentEntityInterface $local_entity,
    array $field_info,
    array $context,
  ) {
    if ($attachments === NULL) {
      return NULL;
    }

    if (!is_array($attachments)) {
      throw new \InvalidArgumentException(sprintf(
        'An array containing the attachments in the format provided by Mailgun was expected, "%s" given.',
        gettype($attachments)
      ));
    }

    $attachments = array_filter(
      $attachments,
      [$this, 'hasPermittedContentType'],
    );
    if (count($attachments) === 0) {
      return NULL;
    }

    $directory = $this->prepareDirectory($local_entity, $field_info);
    $content_id_map = (array) $remote_entity->{'contentIdMap'} ?? [];
    $transformed_attachments = array_values(array_filter(array_map(
      function ($attachment) use (
        $directory,
        $content_id_map,
        $context,
        $field_info,
      ) {
        return $this->transformAttachment(
          $attachment,
          $directory,
          $content_id_map,
          $context,
          $field_info,
        );
      },
      $attachments,
    )));
    return count($transformed_attachments) === 0 ?
      NULL :
      $transformed_attachments;
  }

  /**
   * Transforms a Mailgun attachment to a file entity.
   *
   * @param \stdClass $attachment
   *   The attachment, in the format provided by Mailgun.
   * @param string $directory
   *   The directory where created files should be saved at in the format
   *   "{uri_scheme}://{file_directory}" as expected by
   *   `\Drupal\file\FileRepositoryInterface::writeData()`.
   * @param array $content_id_map
   *   The content ID map, with its items in the format provided by Mailgun but
   *   the wrapper object converted to an array.
   * @param array $context
   *   An associative array containing context about the operation being run.
   * @param array $field_info
   *   An associative array containing information about the field or computed
   *   value being calculated.
   *
   * @return array|null
   *   An array containing the ID of the file entity that was created for the
   *   attachment keyed by `target_id` - as expected by entity reference fields,
   *   or `NULL` if the attachment was not found.
   * phpcs:disable
   * @I Support returning the file entity itself
   *    type     : feature
   *    priority : normal
   *    labels   : attachment
   * @I Support including the content ID in the transformed value
   *    type     : feature
   *    priority : normal
   *    labels   : attachment
   * phpcs:enable
   */
  protected function transformAttachment(
    object $attachment,
    string $directory,
    array $content_id_map,
    array $context,
    array $field_info,
  ): array|null {
    if (!is_object($attachment)) {
      throw new \InvalidArgumentException(sprintf(
        'An object containing the attachment in the format provided by Mailgun was expected, "%s" given.',
        gettype($attachment),
      ));
    }

    $file_contents = $this->getAttachmentContents(
      $attachment,
      $context,
      $field_info,
    );
    if ($file_contents === NULL) {
      return NULL;
    }

    $file = $this->fileRepository->writeData(
      $file_contents,
      "{$directory}/{$attachment->name}",
    );
    // By default, we save the files to be temporary because at this point the
    // message has not been completed yet. A terminate subscriber should be
    // setting those as permanent files after the import of the message has been
    // successful.
    // phpcs:disable
    // @I Add configuration for setting the files as permanent
    //    type     : improvement
    //    priority : normal
    //    labels   : attachment
    // phpcs:enable
    $file->setTemporary();
    $file->save();

    $transformed_attachment = $this->formatTransformedValue(
      $attachment,
      $file,
      $content_id_map,
    );

    $violations = $file->validate();
    if ($violations->count() === 0) {
      return $transformed_attachment;
    }
    if ($this->configuration['skip_failed_attachments'] === TRUE) {
      // phpcs:disable
      // @I Remove files for skipped attachments
      //    type     : bug
      //    priority : normal
      //    labels   : attachment
      //    notes    : The file has already been created so it will remain
      //               orphan without being used anywhere.
      // @I Throw `SkipFieldException` if we are not in computed value
      //    type     : improvement
      //    priority : normal
      //    labels   : attachment
      //    notes    : If we are in a computed value calculation chain we cannot
      //               skip the item; doing so would cause an error due to the
      //               computed value missing in the field that requires it. If,
      //               however, we are in a field calculation we can instruct
      //               the field manager to skip the field by raising a
      //               `SkipFieldException` exception.
      // phpcs:enable
      return NULL;
    }

    foreach ($violations as $violation) {
      $message = sprintf(
        'Failed to validate the file created for attachment "%s" with message "%s".',
        json_encode($attachment),
        (string) $violation->getMessage(),
      );

      throw new FieldImportException($message);
    }

    return $transformed_attachment;
  }

  /**
   * Returns whether the given attachment is of the permitted ones.
   *
   * @param \stdClass $attachment
   *   The attachment, in the format provided by Mailgun.
   *
   * @return bool
   *   `TRUE` if the attachment is permitted, `FALSE` otherwise.
   */
  protected function hasPermittedContentType(object $attachment): bool {
    $settings = $this->configuration['content_types'];
    if (empty($settings)) {
      return TRUE;
    }

    $includes = $settings['include'] ?? [];
    if ($includes && in_array($attachment->{'content-type'}, $includes, TRUE)) {
      return TRUE;
    }

    $excludes = $settings['exclude'] ?? [];
    if ($excludes && !in_array($attachment->{'content-type'}, $excludes, TRUE)) {
      return TRUE;
    }

    // In `validateConfiguration` we make sure that we have either includes, or
    // excludes, or none. If we are here, we have includes and the check failed
    // or we have excludes and the check failed.
    return FALSE;
  }

  /**
   * Prepares and returns the directory where the files should be saved.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $local_entity
   *   The local entity.
   * @param array $field_info
   *   An array containing the field mapping info.
   *
   * @return string
   *   The directory in the format "{uri_scheme}://{file_directory}" as expected
   *   by `\Drupal\file\FileRepositoryInterface::writeData()`.
   *
   * @throws \Drupal\entity_sync\Exception\InvalidConfigurationException
   *   When the target field does not define a valid URL scheme.
   * @throws \Drupal\Core\File\Exception\DirectoryNotReadyException
   *   When the target directory cannot be properly created or configured for
   *   storing the files.
   * phpcs:disable
   * @I Support configuring uri_scheme/file_directory
   *    type     : improvement
   *    priority : normal
   *    labels   : attachment
   *    notes    : That will be useful for using the transformer in computed
   *               values where the machine name does not correspond to an
   *               entity field.
   * @I Support tokens in the file directory
   *    type     : bug
   *    priority : normal
   *    labels   : attachment
   * phpcs:enable
   */
  protected function prepareDirectory(
    ContentEntityInterface $local_entity,
    array $field_info,
  ): string {
    $entity_type_id = NULL;
    $bundle = NULL;
    $field_name = NULL;

    switch ($this->configuration['storage']['source']) {
      case 'target_field':
        $entity_type_id = $local_entity->getEntityTypeId();
        $bundle = $local_entity->bundle();
        $field_name = $field_info['machine_name'];
        break;

      case 'config':
        $entity_type_id = $this->configuration['storage']['entity_type_id'];
        $bundle = $this->configuration['storage']['bundle'];
        $field_name = $this->configuration['storage']['field_name'];
        break;

      default:
        throw new InvalidConfigurationException(sprintf(
          'Unsupported storage configuration option "%s", "target_field" and "config" are supported.',
          $this->configuration['storage']['source'],
        ));
    }

    $definitions = $this->entityFieldManager->getFieldDefinitions(
      $entity_type_id,
      $bundle,
    );
    $definition = $definitions[$field_name];
    $uri_scheme = $definition->getSetting('uri_scheme');
    $file_directory = PlainTextOutput::renderFromHtml($this->token->replace(
      $definition->getSetting('file_directory') ?? '',
      [],
    ));
    if (empty($uri_scheme)) {
      throw new InvalidConfigurationException(
        'No URI scheme setting found for the field.'
      );
    }

    $directory = "{$uri_scheme}://{$file_directory}";
    $directory_exists = $this->fileSystem->prepareDirectory(
      $directory,
      FileSystemInterface::CREATE_DIRECTORY
    );
    if (!$directory_exists) {
      throw new DirectoryNotReadyException(sprintf(
        'The attachment could not be downloaded because the destination directory "%s"is not properly configured. This may be caused by a problem with permissions.',
        $directory
      ));
    }

    return $directory;
  }

  /**
   * Fetches from Mailgun and returns the contents of the given attachment.
   *
   * @param \stdClass $attachment
   *   The attachment object, in the format provided by Mailgun.
   * @param array $context
   *   An associative array containing context about the operation being run.
   * @param array $field_info
   *   An associative array containing information about the field or computed
   *   value being calculated.
   *
   * @return string|null
   *   The contents of the attachment, or `NULL` if the attachment is not found
   *   and the `skip_missing_attachments` in configuration is set to `TRUE`.
   *
   * @throws \RuntimeException
   *   When the Mailgun API key is not available in the operation context.
   * @throws \RuntimeException
   *   When the attachment is not found and the `skip_missing_attachments` in
   *   configuration is set to `FALSE`.
   */
  protected function getAttachmentContents(
    object $attachment,
    array $context,
    array $field_info,
  ): string|null {
    if (!isset($context['entity_sync_mailgun']['api_key'])) {
      throw new \RuntimeException(sprintf(
        'The Mailgun API key is required to be passed in the context so that it
         is available in the "%s" field transformer.',
        __CLASS__
      ));
    }

    // The API settings should be configured at the synchronization/operation
    // type level, rather than in the field transformer configuration
    // settings. We therefore require them to be made available in the context -
    // that's the only way to get them in here.
    $arguments = [$context['entity_sync_mailgun']['api_key']];
    if (array_key_exists('endpoint', $context['entity_sync_mailgun'])) {
      $arguments[] = $context['entity_sync_mailgun']['endpoint'];
    }

    try {
      return (string) Mailgun::create(...$arguments)
        ->attachment()
        ->show($attachment->url)
        ->getBody();
    }
    catch (HttpClientException $exception) {
      if ($exception->getCode() !== 404) {
        throw $exception;
      }

      // phpcs:disable
      // @I Throw `SkipFieldException` if we are not in computed value
      //    type     : improvement
      //    priority : normal
      //    labels   : attachment
      //    notes    : If we are in a computed value calculation chain we cannot
      //               skip the item; doing so would cause an error due to the
      //               computed value missing in the field that requires it. If,
      //               however, we are in a field calculation we can instruct
      //               the field manager to skip the field by raising a
      //               `SkipFieldException` exception.
      // phpcs:enable
      if ($this->configuration['skip_missing_attachments'] === TRUE) {
        return NULL;
      }

      throw new \RuntimeException(
        'Attachment not found. Mailgun email attachments are available for retrieval for a limited time only after an email is sent.'
      );
    }
  }

  /**
   * Format the transformed value according to the configured format.
   *
   * @param \stdClass $attachment
   *   The attachment object, in the format provided by Mailgun.
   * @param \Drupal\file\FileInterface $file
   *   The file that was created for the attachment.
   * @param array $content_id_map
   *   The content ID map, with its items in the format provided by Mailgun but
   *   the wrapper object converted to an array.
   *
   * @return array
   *   The transformed value in the format requested in the plugin
   *   configuration.
   */
  protected function formatTransformedValue(
    object $attachment,
    FileInterface $file,
    array $content_id_map,
  ): array {
    if ($this->configuration['format'] === 'entity_reference') {
      return ['target_id' => $file->id()];
    }

    if ($this->configuration['format'] !== 'content_id') {
      throw new InvalidConfigurationException(sprintf(
        'Unknown file attachment format "%s", "entity_reference" and "content_id" are supported.',
        $this->configuration['format']
      ));
    }

    $content_id_map = array_filter(
      $content_id_map,
      function ($map_item) use ($attachment) {
        return $map_item->url === $attachment->url;
      }
    );
    if (count($content_id_map) !== 1) {
      return [$file->id() => FALSE];
    }

    // The value is wrapped in '<>'.
    $content_id = trim(key($content_id_map), '<>');
    return [$file->id() => $content_id];
  }

}
