<?php

declare(strict_types=1);

namespace Drupal\entity_sync_mailgun\Plugin\EntitySync\FieldTransformer;

// Drupal modules.
use Drupal\entity_sync\FieldTransformer\PluginBase;
use Drupal\file\FileStorageInterface;
// Drupal core.
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
// Third-party libraries.
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\Validator\ValidatorInterface;

/**
 * Transformer that replaces Mailgun inline image placeholders.
 *
 * Email messages contain inline image tags with the content ID of the image
 * attachment in the source attribute e.g. `<img src="cid:f_lyw4u2352">`. We
 * replace these with the URL to the file created for the attachment. We also
 * add the `data-entity-type` and `data-entity-uuid` attributes so that the
 * `editor_file_reference` filter plugin can be taken advantage of.
 *
 * Supported configuration properties:
 * - content_id_map_keys: (array, required) An array containing the keys of the
 *   computed values that contain content ID maps in the format returned by the
 *   `mailgun_file_attachments` field transformer.
 *
 * @EntitySyncFieldTransformer(
 *   id = "mailgun_inline_file_images"
 * )
 */
class InlineFileImages extends PluginBase {

  /**
   * The file storage.
   */
  protected FileStorageInterface $fileStorage;

  /**
   * Constructs a new InlineFileImages object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Symfony\Component\Validator\Validator\ValidatorInterface $constraint_validator
   *   The constraint validator.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $constraint_manager
   *   The constraint plugin manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    ValidatorInterface $constraint_validator,
    PluginManagerInterface $constraint_manager,
    protected EntityTypeManagerInterface $entityTypeManager,
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $constraint_validator,
      $constraint_manager,
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    ValidatorInterface $constraint_validator,
    array $configuration,
    $plugin_id,
    $plugin_definition,
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $constraint_validator,
      $container->get('plugin.manager.entity_sync_validation_constraint'),
      $container->get('entity_type.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function validateConfiguration() {
    $this->validateConfigurationRequiredProperties([
      'content_id_map_keys',
    ]);
  }

  /**
   * {@inheritdoc}
   */
  protected function transformImportedValue(
    $html,
    \stdClass $remote_entity,
    ?ContentEntityInterface $local_entity,
    array $field_info,
    array $context,
  ) {
    if ($html === NULL) {
      return NULL;
    }

    if (!is_string($html)) {
      throw new \InvalidArgumentException(sprintf(
        'The value must be a string, "%s" given',
        gettype($html),
      ));
    }

    $files = $this->loadFiles($context);
    if (count($files) === 0) {
      return $html;
    }

    $dom = Html::load($html);
    $xpath = new \DOMXPath($dom);

    foreach ($xpath->query('//img[@src]') as $element) {
      $src = $element->getAttribute('src');
      if (!str_starts_with($src, 'cid:')) {
        continue;
      }

      $src_parts = explode(':', $src);
      if (count($src_parts) !== 2) {
        continue;
      }

      $content_id = $src_parts[1];
      if (!isset($files[$content_id])) {
        // phpcs:disable
        // @I Log an error if the file is not found
        //    type     : bug
        //    priority : low
        //    labels   : attachment, html
        // phpcs:enable
        continue;
      }

      $file = $files[$content_id];
      $element->setAttribute('src', $file->createFileUrl());
      $element->setAttribute('data-entity-type', 'file');
      $element->setAttribute('data-entity-uuid', $file->uuid());
    }

    return Html::serialize($dom);
  }

  /**
   * Returns an array containing the file entities.
   *
   * The context must contain one or more set of content ID maps as computed
   * values in the format returned by the `mailgun_file_attachments` field
   * transformer.
   *
   * @param array $context
   *   An associative array containing context about the operation being run.
   *
   * @return array
   *   An array containing the file entities, keyed by the content ID of the
   *   attachment for which the entity was created.
   */
  protected function loadFiles(array $context): array {
    $content_id_map = [];
    foreach ($this->configuration['content_id_map_keys'] as $key) {
      if (!array_key_exists($key, $context['computed_values'] ?? [])) {
        throw new \RuntimeException(sprintf(
          'No computed value available for key "%s".',
          $key,
        ));
      }
      if ($context['computed_values'][$key] === NULL) {
        continue;
      }
      $content_id_map = array_merge(
        $content_id_map,
        $context['computed_values'][$key],
      );
    }

    $files = [];
    foreach ($content_id_map as $item) {
      $file_id = key($item);
      $content_id = current($item);
      $file = $this->fileStorage()->load($file_id);
      if (!$file) {
        throw new \RuntimeException(sprintf(
          'File entity with ID "%s" not found for attachment with content ID "%s".',
          $file_id,
          $content_id,
        ));
      }
      if (!isset($files[$content_id])) {
        $files[$content_id] = $file;
        continue;
      }

      if ($files[$content_id]->id() == $file_id) {
        continue;
      }

      throw new \RuntimeException(sprintf(
        'Attachment with content ID "%s" imported multiple times as file entities with IDs %s.',
        $content_id,
        implode(', ', [$files[$content_id]->id(), $file_id]),
      ));
    }

    return $files;
  }

  /**
   * Returns the file storage.
   *
   * @return \Drupal\file\FileStorageInterface
   *   The file storage.
   */
  protected function fileStorage(): FileStorageInterface {
    if (isset($this->fileStorage)) {
      return $this->fileStorage;
    }

    $this->fileStorage = $this->entityTypeManager->getStorage('file');
    return $this->fileStorage;
  }

}
