<?php

declare(strict_types=1);

namespace Drupal\entity_sync_mailgun\Plugin\EntitySync\FieldTransformer;

use Drupal\entity_sync\FieldTransformer\PluginBase;
use Drupal\Core\Entity\ContentEntityInterface;

/**
 * Transformer that merges file attachments from multiple fields.
 *
 * Currently, this applies to exports only. It allows merging files coming from
 * multiple fields into one array of attachments for passing them to
 * Mailgun. The input value needs to be an array of arrays, each one of which
 * must be an array containing the path to the file as the first element.
 *
 * For example, given the following input value:
 * @code
 * $value = [
 *   'field_files' => [
 *     0 => [
 *       0 => '/path/to/file/my_file.pdf',
 *     ],
 *   ],
 *   'field_images' => [
 *     0 => [
 *       0 => '/path/to/file/an_image.png',
 *     ],
 *     1 => [
 *       0 => '/path/to/file/another_image.jpg',
 *     ],
 *   ],
 * ];
 * @endcode
 * The transformed value will be:
 * @code
 * $value = [
 *   0 => [
 *     0 => '/path/to/file/my_file.pdf',
 *   ],
 *   1 => [
 *     0 => '/path/to/file/an_image.png',
 *   ],
 *   2 => [
 *     0 => '/path/to/file/another_image.jpg',
 *   ],
 * ];
 * @endcode
 *
 * @EntitySyncFieldTransformer(
 *   id = "mailgun_merge_file_attachments"
 * )
 */
class MergeFileAttachments extends PluginBase {

  /**
   * {@inheritdoc}
   */
  protected function transformExportedValue(
    $value,
    ContentEntityInterface $local_entity,
    $remote_entity_id,
    array $field_info,
    array $context,
  ) {
    if ($value === NULL) {
      return NULL;
    }

    if (!is_array($value)) {
      throw new \RuntimeException(sprintf(
        'The value being converted must be an array, "%s" given.',
        gettype($value)
      ));
    }

    if (count($value) === 0) {
      return NULL;
    }

    $files = [];
    foreach ($value as $file_group) {
      if (empty($file_group)) {
        continue;
      }
      $files = array_merge($files, $file_group);
    }

    if (count($files) === 0) {
      return NULL;
    }

    return $files;
  }

}
