<?php

declare(strict_types=1);

namespace Drupal\entity_sync_mailgun\Plugin\EntitySync\FieldTransformer;

use Drupal\entity_sync\FieldTransformer\PluginBase;
use Drupal\Core\Entity\ContentEntityInterface;

/**
 * Transformer that imports the time from a message's headers.
 *
 * @EntitySyncFieldTransformer(
 *   id = "mailgun_time"
 * )
 * phpcs:disable
 * @I Support option to fall back to the current request time
 *    type     : improvement
 *    priority : low
 *    labels   : import
 *    notes    : There could be cases where we cannot parse correctly the date
 *               headers, or they may not be available. In such cases
 *               applications may prefer to use the time that the message is
 *               imported, the accuracy of which depending on the way the the
 *               imports are run it acceptable for the purposed of the
 *               application. Some applications may also choose to store the raw
 *               message and parse it again later to correct message times.
 * phpcs:enable
 */
class Time extends PluginBase {

  /**
   * {@inheritdoc}
   */
  protected function transformImportedValue(
    $headers,
    \stdClass $remote_entity,
    ?ContentEntityInterface $local_entity,
    array $field_info,
    array $context,
  ) {
    if ($headers === NULL) {
      return NULL;
    }

    if (!is_array($headers)) {
      throw new \InvalidArgumentException(sprintf(
        'An array containing the headers in the format provided by Mailgun was expected, "%s" given.',
        gettype($headers)
      ));
    }

    if (count($headers) === 0) {
      return NULL;
    }

    $header_values = $this->getDateHeaderValues($headers);
    if (count($header_values) === 0) {
      return NULL;
    }

    foreach ($header_values as $date) {
      $time = $this->dateStringToTimestamp($date);
      if ($time !== FALSE) {
        return $time;
      }
    }

    return NULL;
  }

  /**
   * Returns the values from all headers containing the message's time.
   *
   * The time of the email is contained in different formats, with values that
   * do not necessarily match. This might be because different email servers
   * (sender, mailgun) add their own time headers ("X-Received", "Received")
   * that seems to be the time they received the message while the "Date" header
   * seems to be the time that the message was created.
   *
   * Not all of these headers exist though. We therefore fetch all available
   * values, prioritizing the "Date" header first and then the other ones. The
   * first value that can be parsed to a timestamp should be used by the caller.
   *
   * @param array $headers
   *   An array containing the headers of the message, in the format provided by
   *   Mailgun.
   *
   * @return array
   *   An array containing the values of headers containing the message time.
   */
  protected function getDateHeaderValues(array $headers): array {
    // Some headers (e.g. "Received") can exist multiple times. Get the values
    // grouped by header name removing empty results, resulting in a
    // two-dimensional array.
    $grouped_header_values = array_map(
      function ($header) use ($headers) {
        return array_map(
          fn ($h) => $h[1],
          array_filter($headers, fn ($h) => $header === $h[0]),
        );
      },
      ['Date', 'X-Received', 'Received'],
    );
    $grouped_header_values = array_filter(
      $grouped_header_values,
      fn ($h) => count($h) !== 0,
    );

    // Flatten the array so that it contains directly the date values.
    $header_values = [];
    foreach ($grouped_header_values as $header_values_group) {
      $header_values = array_merge($header_values, $header_values_group);
    }

    return $header_values;
  }

  /**
   * Converts the given date to a Unix timestamp.
   *
   * @param string $date
   *   The date to convert.
   *
   * @return int|false
   *   The Unix timestamp, or `FALSE` if the format is unknown or cannot be
   *   parsed by `strtotime()`.
   */
  protected function dateStringToTimestamp(string $date): int|false {
    $date = trim($date);

    // Some headers contain the date string after ';'.
    $date_parts = explode(';', $date);
    if (count($date_parts) > 2) {
      return FALSE;
    }

    $date = count($date_parts) === 2 ? $date_parts[1] : $date_parts[0];

    // Some headers unnecessarily or even incorrectly contain the timezone in
    // parenthesis at the end, which makes it incompatible for `strtotime()`.
    $date_subparts = explode('(', $date);
    if (count($date_subparts) > 2) {
      return FALSE;
    }

    return strtotime(trim($date_subparts[0]));
  }

}
